"""Provide functions to add/find contract checkers."""
import contextvars
import functools
import inspect
from typing import (
    Callable,
    Any,
    Iterable,
    Optional,
    Tuple,
    List,
    Mapping,
    MutableMapping,
    Dict,
    cast,
    Set,
)

import icontract._represent
from icontract._globals import CallableT, ClassT
from icontract._types import Contract, Snapshot
from icontract.errors import ViolationError

# pylint does not play with typing.Mapping.
# pylint: disable=unsubscriptable-object
# pylint: disable=raising-bad-type


def _walk_decorator_stack(func: CallableT) -> Iterable["CallableT"]:
    """
    Iterate through the stack of decorated functions until the original function.

    Assume that all decorators used functools.update_wrapper.
    """
    while hasattr(func, "__wrapped__"):
        yield func

        func = getattr(func, "__wrapped__")

    yield func


def find_checker(func: CallableT) -> Optional[CallableT]:
    """Iterate through the decorator stack till we find the contract checker."""
    contract_checker = None  # type: Optional[CallableT]
    for a_wrapper in _walk_decorator_stack(func):
        if hasattr(a_wrapper, "__preconditions__") or hasattr(
            a_wrapper, "__postconditions__"
        ):
            contract_checker = a_wrapper

    return contract_checker


def kwargs_from_call(
    param_names: List[str],
    kwdefaults: Dict[str, Any],
    args: Tuple[Any, ...],
    kwargs: Dict[str, Any],
) -> MutableMapping[str, Any]:
    """
    Inspect the input values received at the wrapper for the actual function call.

    :param param_names: parameter (*i.e.* argument) names of the original (decorated) function
    :param kwdefaults: default argument values of the original function
    :param args: arguments supplied to the call
    :param kwargs: keyword arguments supplied to the call
    :return: resolved arguments as they would be passed to the function
    """
    # (Marko Ristin, 2020-12-01)
    # Insert _ARGS and _KWARGS preemptively even if they are not needed by any contract.
    # This makes the code logic much simpler since we do not explicitly check if a contract would
    # need them, though it might incur a subtle computational overhead
    # (*e.g.*, when the contracts do not need them or don't use any argument at all).
    # We need to have a concrete issue where profiling helps us determine if this is a real
    # bottleneck or not and not optimize for no real benefit.
    resolved_kwargs = {"_ARGS": args, "_KWARGS": kwargs}

    # Set the default argument values as condition parameters.
    for param_name, param_value in kwdefaults.items():
        resolved_kwargs[param_name] = param_value

    # Override the defaults with the values actually supplied to the function.
    for i, func_arg in enumerate(args):
        if i < len(param_names):
            resolved_kwargs[param_names[i]] = func_arg
        else:
            # Silently ignore call arguments that were not specified in the function.
            # This way we let the underlying decorated function raise the exception
            # instead of frankensteining the exception here.

            # It seems that this line can not be covered,
            # see https://github.com/nedbat/coveragepy/issues/1041.
            # The branch was covered manually in ``tests.test_checkers``.
            pass  # pragma: no cover

    for key, val in kwargs.items():
        resolved_kwargs[key] = val

    return resolved_kwargs


def not_check(check: Any, contract: Contract) -> bool:
    """
    Negate the check value of a condition and capture missing boolyness (*e.g.*, when check is a numpy array).

    :param check: value of the evaluated condition
    :param contract: corresponding to the check
    :return: negated check
    :raise: ValueError if the check could not be negated
    """
    try:
        return not check
    except Exception as err:  # pylint: disable=broad-except
        msg_parts = []  # type: List[str]
        if contract.location is not None:
            msg_parts.append("{}:\n".format(contract.location))

        msg_parts.append("Failed to negate the evaluation of the condition.")

        raise ValueError("".join(msg_parts)) from err


def select_condition_kwargs(
    contract: Contract, resolved_kwargs: Mapping[str, Any]
) -> Mapping[str, Any]:
    """
    Select the keyword arguments that are used by the contract.

    :param contract: contract to be verified
    :param resolved_kwargs:
        resolved keyword arguments of the call (including the default argument values of the decorated function)
    :return: a subset of resolved_kwargs
    """
    # Check that all arguments to the condition function have been set.
    missing_args = [
        arg_name
        for arg_name in contract.mandatory_args
        if arg_name not in resolved_kwargs
    ]
    if missing_args:
        msg_parts = []  # type: List[str]
        if contract.location is not None:
            msg_parts.append("{}:\n".format(contract.location))

        msg_parts.append(
            (
                "The argument(s) of the contract condition have not been set: {}. "
                "Does the original function define them? Did you supply them in the call?"
            ).format(missing_args)
        )

        if "OLD" in missing_args:
            msg_parts.append(
                " Did you decorate the function with a snapshot to capture OLD values?"
            )

        raise TypeError("".join(msg_parts))

    condition_kwargs = {
        arg_name: value
        for arg_name, value in resolved_kwargs.items()
        if arg_name in contract.condition_arg_set
    }

    return condition_kwargs


def _assert_no_invalid_kwargs(kwargs: Any) -> Optional[TypeError]:
    """Check that kwargs of a function contain no unexpected arguments."""
    if "_ARGS" in kwargs:
        return TypeError(
            'The arguments of the function call include "_ARGS" which is '
            "a placeholder for positional arguments in a condition."
        )

    if "_KWARGS" in kwargs:
        return TypeError(
            'The arguments of the function call include "_KWARGS" which is '
            "a placeholder for keyword arguments in a condition."
        )

    return None


def _unpack_pre_snap_posts(
    wrapper: CallableT,
) -> Tuple[List[List[Contract]], List[Snapshot], List[Contract]]:
    """Retrieve the preconditions, snapshots and postconditions defined for the given wrapper checker."""
    preconditions = getattr(wrapper, "__preconditions__")  # type: List[List[Contract]]
    snapshots = getattr(wrapper, "__postcondition_snapshots__")  # type: List[Snapshot]
    postconditions = getattr(wrapper, "__postconditions__")  # type: List[Contract]

    return preconditions, snapshots, postconditions


def _assert_resolved_kwargs_valid(
    postconditions: List[Contract], resolved_kwargs: Mapping[str, Any]
) -> Optional[TypeError]:
    """Check that the resolved kwargs of a decorated function are valid."""
    if postconditions:
        if "result" in resolved_kwargs:
            return TypeError(
                "Unexpected argument 'result' in a function decorated with postconditions."
            )

        if "OLD" in resolved_kwargs:
            return TypeError(
                "Unexpected argument 'OLD' in a function decorated with postconditions."
            )

    return None


def _create_violation_error(
    contract: Contract, resolved_kwargs: Mapping[str, Any]
) -> BaseException:
    """Create the violation error based on the violated contract."""
    exception = None  # type: Optional[BaseException]

    if contract.error is None:
        try:
            msg = icontract._represent.generate_message(
                contract=contract, resolved_kwargs=resolved_kwargs
            )
        except Exception as err:
            parts = ["Failed to recompute the values of the contract condition:\n"]
            if contract.location is not None:
                parts.append("{}:\n".format(contract.location))

            if contract.description is not None:
                parts.append("{}: ".format(contract.description))

            parts.append(
                icontract._represent.represent_condition(condition=contract.condition)
            )

            raise RuntimeError("".join(parts)) from err

        exception = ViolationError(msg)
    elif inspect.ismethod(contract.error) or inspect.isfunction(contract.error):
        assert (
            contract.error_arg_set is not None
        ), "Expected error_arg_set non-None if contract.error a function."
        assert (
            contract.error_args is not None
        ), "Expected error_args non-None if contract.error a function."

        error_kwargs = select_error_kwargs(
            contract=contract, resolved_kwargs=resolved_kwargs
        )

        exception = cast(BaseException, contract.error(**error_kwargs))

        if not isinstance(exception, BaseException):
            raise TypeError(
                "The exception returned by the contract's error {} does not inherit from BaseException.".format(
                    contract.error
                )
            )
    elif isinstance(contract.error, type):
        if not issubclass(contract.error, BaseException):
            raise TypeError(
                "The exception class supplied in the contract's error {} is not a subclass of BaseException.".format(
                    contract.error
                )
            )

        msg = icontract._represent.generate_message(
            contract=contract, resolved_kwargs=resolved_kwargs
        )
        exception = contract.error(msg)
    elif isinstance(contract.error, BaseException):
        exception = contract.error
    else:
        raise NotImplementedError(
            (
                "icontract does not know how to handle the error of type {} "
                "(expected a function, a subclass of BaseException or an instance of BaseException)"
            ).format(type(contract.error))
        )

    assert exception is not None
    return exception


async def _assert_preconditions_async(
    preconditions: List[List[Contract]], resolved_kwargs: Mapping[str, Any]
) -> Optional[BaseException]:
    """Assert that the preconditions of an async function hold."""
    exception = None  # type: Optional[BaseException]

    # Assert the preconditions in groups. This is necessary to implement "require else" logic when a class
    # weakens the preconditions of its base class.

    for group in preconditions:
        exception = None

        for contract in group:
            assert (
                exception is None
            ), "No exception as long as pre-condition group is satisfiable."

            condition_kwargs = select_condition_kwargs(
                contract=contract, resolved_kwargs=resolved_kwargs
            )

            if inspect.iscoroutinefunction(contract.condition):
                check = await contract.condition(**condition_kwargs)
            else:
                check_or_coroutine = contract.condition(**condition_kwargs)
                if inspect.iscoroutine(check_or_coroutine):
                    check = await check_or_coroutine
                else:
                    check = check_or_coroutine

            if not_check(check=check, contract=contract):
                exception = _create_violation_error(
                    contract=contract, resolved_kwargs=resolved_kwargs
                )
                break

        # The group of preconditions was satisfied, no need to check the other groups.
        if exception is None:
            break

    return exception


def _assert_preconditions(
    preconditions: List[List[Contract]],
    resolved_kwargs: Mapping[str, Any],
    func: CallableT,
) -> Optional[BaseException]:
    """Assert that the preconditions of a sync function hold."""
    exception = None  # type: Optional[BaseException]

    # Assert the preconditions in groups. This is necessary to implement "require else" logic when a class
    # weakens the preconditions of its base class.

    for group in preconditions:
        exception = None

        for contract in group:
            assert (
                exception is None
            ), "No exception as long as pre-condition group is satisfiable."

            condition_kwargs = select_condition_kwargs(
                contract=contract, resolved_kwargs=resolved_kwargs
            )

            if inspect.iscoroutinefunction(contract.condition):
                raise ValueError(
                    "Unexpected coroutine (async) condition {} for a sync function {}.".format(
                        contract.condition, func
                    )
                )

            check = contract.condition(**condition_kwargs)

            if inspect.iscoroutine(check):
                raise ValueError(
                    "Unexpected coroutine resulting from the condition {} for a sync function {}.".format(
                        contract.condition, func
                    )
                )

            if not_check(check=check, contract=contract):
                exception = _create_violation_error(
                    contract=contract, resolved_kwargs=resolved_kwargs
                )
                break

        # The group of preconditions was satisfied, no need to check the other groups.
        if exception is None:
            break

    return exception


async def _capture_old_async(
    snapshots: List[Snapshot], resolved_kwargs: Mapping[str, Any]
) -> "Old":
    """Capture all snapshots of an async function and return the captured values bundled in an ``Old``."""
    old_as_mapping = dict()  # type: MutableMapping[str, Any]
    for snap in snapshots:
        # This assert is just a last defense.
        # Conflicting snapshot names should have been caught before, either during the decoration or
        # in the meta-class.
        assert (
            snap.name not in old_as_mapping
        ), "Snapshots with the conflicting name: {}"

        capture_kwargs = select_capture_kwargs(
            a_snapshot=snap, resolved_kwargs=resolved_kwargs
        )

        if inspect.iscoroutinefunction(snap.capture):
            old_as_mapping[snap.name] = await snap.capture(**capture_kwargs)
        else:
            captured_or_coroutine = snap.capture(**capture_kwargs)
            if inspect.iscoroutine(captured_or_coroutine):
                captured = await captured_or_coroutine
            else:
                captured = captured_or_coroutine

            old_as_mapping[snap.name] = captured

    return Old(mapping=old_as_mapping)


def _capture_old(
    snapshots: List[Snapshot], resolved_kwargs: Mapping[str, Any], func: CallableT
) -> "Old":
    """Capture all snapshots of a sync function and return the captured values bundled in an ``Old``."""
    old_as_mapping = dict()  # type: MutableMapping[str, Any]
    for snap in snapshots:
        # This assert is just a last defense.
        # Conflicting snapshot names should have been caught before, either during the decoration or
        # in the meta-class.
        assert (
            snap.name not in old_as_mapping
        ), "Snapshots with the conflicting name: {}"

        if inspect.iscoroutinefunction(snap.capture):
            raise ValueError(
                "Unexpected coroutine (async) snapshot capture {} for a sync function {}.".format(
                    snap.capture, func
                )
            )

        capture_kwargs = select_capture_kwargs(
            a_snapshot=snap, resolved_kwargs=resolved_kwargs
        )

        captured = snap.capture(**capture_kwargs)
        if inspect.iscoroutine(captured):
            raise ValueError(
                (
                    "Unexpected coroutine resulting from the snapshot capture {} "
                    "of a sync function {}."
                ).format(snap.capture, func)
            )

        old_as_mapping[snap.name] = captured

    return Old(mapping=old_as_mapping)


async def _assert_postconditions_async(
    postconditions: List[Contract], resolved_kwargs: Mapping[str, Any]
) -> Optional[BaseException]:
    """Assert that the postconditions of an async function hold."""
    assert (
        "result" in resolved_kwargs
    ), "Expected 'result' to be already set in resolved kwargs before calling this function."

    for contract in postconditions:
        condition_kwargs = select_condition_kwargs(
            contract=contract, resolved_kwargs=resolved_kwargs
        )

        if inspect.iscoroutinefunction(contract.condition):
            check = await contract.condition(**condition_kwargs)
        else:
            check_or_coroutine = contract.condition(**condition_kwargs)
            if inspect.iscoroutine(check_or_coroutine):
                check = await check_or_coroutine
            else:
                check = check_or_coroutine

        if not_check(check=check, contract=contract):
            exception = _create_violation_error(
                contract=contract, resolved_kwargs=resolved_kwargs
            )

            return exception

    return None


def _assert_postconditions(
    postconditions: List[Contract], resolved_kwargs: Mapping[str, Any], func: CallableT
) -> Optional[BaseException]:
    """Assert that the postconditions of a sync function hold."""
    assert (
        "result" in resolved_kwargs
    ), "Expected 'result' to be already set in resolved kwargs before calling this function."

    for contract in postconditions:
        if inspect.iscoroutinefunction(contract.condition):
            raise ValueError(
                "Unexpected coroutine (async) condition {} for a sync function {}.".format(
                    contract.condition, func
                )
            )

        condition_kwargs = select_condition_kwargs(
            contract=contract, resolved_kwargs=resolved_kwargs
        )

        check = contract.condition(**condition_kwargs)

        if inspect.iscoroutine(check):
            raise ValueError(
                "Unexpected coroutine resulting from the condition {} for a sync function {}.".format(
                    contract.condition, func
                )
            )

        if not_check(check=check, contract=contract):
            exception = _create_violation_error(
                contract=contract, resolved_kwargs=resolved_kwargs
            )

            return exception

    return None


def _assert_invariant(contract: Contract, instance: Any) -> None:
    """Assert that the contract holds as a class invariant given the instance of the class."""
    if "self" in contract.condition_arg_set:
        check = contract.condition(self=instance)
    else:
        check = contract.condition()

    if not_check(check=check, contract=contract):
        raise _create_violation_error(
            contract=contract, resolved_kwargs={"self": instance}
        )


def select_capture_kwargs(
    a_snapshot: Snapshot, resolved_kwargs: Mapping[str, Any]
) -> Mapping[str, Any]:
    """
    Select the keyword arguments that are used by the snapshot capture.

    :param a_snapshot: snapshot to be captured
    :param resolved_kwargs: resolved keyword arguments (including the default values)
    :return: a subset of resolved_kwargs
    """
    missing_args = [
        arg_name for arg_name in a_snapshot.args if arg_name not in resolved_kwargs
    ]
    if missing_args:
        msg_parts = []
        if a_snapshot.location is not None:
            msg_parts.append("{}:\n".format(a_snapshot.location))

        msg_parts.append(
            (
                "The argument(s) of the snapshot have not been set: {}. "
                "Does the original function define them? Did you supply them in the call?"
            ).format(missing_args)
        )

        raise TypeError("".join(msg_parts))

    return {
        arg_name: arg_value
        for arg_name, arg_value in resolved_kwargs.items()
        if arg_name in a_snapshot.arg_set
    }


def select_error_kwargs(
    contract: Contract, resolved_kwargs: Mapping[str, Any]
) -> Mapping[str, Any]:
    """
    Select the keyword arguments that are used by the error creator of the contract.

    :param contract: contract that was violated and for which we want to generate an error
    :param resolved_kwargs: resolved keyword arguments (including the default values)
    :return: a subset of resolved_kwargs
    """
    assert contract.error_arg_set is not None
    assert contract.error_args is not None

    error_kwargs = {
        arg_name: value
        for arg_name, value in resolved_kwargs.items()
        if arg_name in contract.error_arg_set
    }

    missing_args = [
        arg_name for arg_name in contract.error_args if arg_name not in resolved_kwargs
    ]
    if missing_args:
        msg_parts = []  # type: List[str]
        if contract.location is not None:
            msg_parts.append("{}:\n".format(contract.location))

        msg_parts.append(
            (
                "The argument(s) of the contract error have not been set: {}. "
                "Does the original function define them? Did you supply them in the call?"
            ).format(missing_args)
        )

        raise TypeError("".join(msg_parts))

    return error_kwargs


class Old:
    """
    Represent argument values before the function invocation.

    Recipe taken from http://code.activestate.com/recipes/52308-the-simple-but-handy-collector-of-a-bunch-of-named/
    """

    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """Update the ``__dict__`` with the given mapping."""
        self.__dict__.update(mapping)

    def __getattr__(self, item: str) -> Any:
        """Raise an error as this ``item`` should not be in the ``__dict__``."""
        raise AttributeError(
            "The snapshot with the name {!r} is not available in the OLD of a postcondition. "
            "Have you decorated the function with a corresponding snapshot decorator?".format(
                item
            )
        )

    def __repr__(self) -> str:
        """Represent the old values with a string literal as user is unaware of the class."""
        return "a bunch of OLD values"


def resolve_kwdefaults(sign: inspect.Signature) -> Dict[str, Any]:
    """Resolve default values for the function arguments based on its signature."""
    kwdefaults = dict()  # type: Dict[str, Any]

    # Add to the defaults all the values that are needed by the contracts.
    for param in sign.parameters.values():
        if param.default != inspect.Parameter.empty:
            kwdefaults[param.name] = param.default

    return kwdefaults


# This flag is used to avoid recursively checking contracts for the same function or instance while
# contract checking is already in progress.
#
# The key refers to the id() of the function (preconditions and postconditions) or instance (invariants).
_IN_PROGRESS = contextvars.ContextVar(
    "_IN_PROGRESS", default=None
)  # type: contextvars.ContextVar[Optional[Set[int]]]


def decorate_with_checker(func: CallableT) -> CallableT:
    """Decorate the function with a checker that verifies the preconditions and postconditions."""
    assert not hasattr(
        func, "__preconditions__"
    ), "Expected func to have no list of preconditions (there should be only a single contract checker per function)."

    assert not hasattr(
        func, "__postconditions__"
    ), "Expected func to have no list of postconditions (there should be only a single contract checker per function)."

    assert not hasattr(func, "__postcondition_snapshots__"), (
        "Expected func to have no list of postcondition snapshots (there should be only a single contract checker "
        "per function)."
    )

    sign = inspect.signature(func)
    if "_ARGS" in sign.parameters:
        raise TypeError(
            'The arguments of the function to be decorated with a contract checker include "_ARGS" which is '
            "a reserved placeholder for positional arguments in the condition."
        )

    if "_KWARGS" in sign.parameters:
        raise TypeError(
            'The arguments of the function to be decorated with a contract checker include "_KWARGS" which is '
            "a reserved placeholder for keyword arguments in the condition."
        )

    param_names = list(sign.parameters.keys())

    # Determine the default argument values
    kwdefaults = resolve_kwdefaults(sign=sign)

    id_func = id(func)

    # (mristin, 2021-02-16)
    # Admittedly, this branching on sync/async is absolutely monstrous.
    # However, I couldn't find out an easier way to refactor the code so that it supports async.
    # Python expects us to explicitly colour functions as sync/async so we can not just put in an if-statement and
    # introduce an "await".
    #
    # The two wrappers need to be manually maintained in parallel.
    # Whenever you make a change, please inspect manually that both sync and async code exercises equivalent behavior.
    # For example, copy/paste the two blocks of code in separate files and perform a diff.

    if inspect.iscoroutinefunction(func):

        async def wrapper(*args, **kwargs):  # type: ignore
            """Wrap func by checking the preconditions and postconditions."""
            kwargs_error = _assert_no_invalid_kwargs(kwargs)
            if kwargs_error:
                raise kwargs_error

            # We need to create a new in-progress set if it is None as the ``ContextVar`` does not accept
            # a factory function for the default argument. If we didn't do this, and simply set an empty
            # set as the default, ``ContextVar`` would always point to the same set by copying the default
            # by reference.
            in_progress = _IN_PROGRESS.get()
            if in_progress is None:
                in_progress = set()
                _IN_PROGRESS.set(in_progress)

            # Use try-finally instead of ExitStack for performance.
            try:
                # If the wrapper is already checking the contracts for the wrapped function, avoid a recursive loop
                # by skipping any subsequent contract checks for the same function.
                if id_func in in_progress:
                    return await func(*args, **kwargs)

                in_progress.add(id_func)

                (preconditions, snapshots, postconditions) = _unpack_pre_snap_posts(
                    wrapper
                )

                resolved_kwargs = kwargs_from_call(
                    param_names=param_names,
                    kwdefaults=kwdefaults,
                    args=args,
                    kwargs=kwargs,
                )

                type_error = _assert_resolved_kwargs_valid(
                    postconditions, resolved_kwargs
                )
                if type_error:
                    raise type_error

                violation_error = await _assert_preconditions_async(
                    preconditions=preconditions, resolved_kwargs=resolved_kwargs
                )
                if violation_error:
                    raise violation_error

                # Capture the snapshots
                if postconditions and snapshots:
                    resolved_kwargs["OLD"] = await _capture_old_async(
                        snapshots=snapshots, resolved_kwargs=resolved_kwargs
                    )

                # Ideally, we would catch any exception here and strip the checkers from the traceback.
                # Unfortunately, this can not be done in Python 3, see
                # https://stackoverflow.com/questions/44813333/how-can-i-elide-a-function-wrapper-from-the-traceback-in-python-3
                result = await func(*args, **kwargs)

                if postconditions:
                    resolved_kwargs["result"] = result

                    violation_error = await _assert_postconditions_async(
                        postconditions=postconditions, resolved_kwargs=resolved_kwargs
                    )
                    if violation_error:
                        raise violation_error

                return result
            finally:
                in_progress.discard(id_func)

    else:

        def wrapper(*args, **kwargs):  # type: ignore
            """Wrap func by checking the preconditions and postconditions."""
            kwargs_error = _assert_no_invalid_kwargs(kwargs)
            if kwargs_error:
                raise kwargs_error

            # We need to create a new in-progress set if it is None as the ``ContextVar`` does not accept
            # a factory function for the default argument. If we didn't do this, and simply set an empty
            # set as the default, ``ContextVar`` would always point to the same set by copying the default
            # by reference.
            in_progress = _IN_PROGRESS.get()
            if in_progress is None:
                in_progress = set()
                _IN_PROGRESS.set(in_progress)

            # Use try-finally instead of ExitStack for performance.
            try:
                # If the wrapper is already checking the contracts for the wrapped function, avoid a recursive loop
                # by skipping any subsequent contract checks for the same function.
                if id_func in in_progress:
                    return func(*args, **kwargs)

                in_progress.add(id_func)

                (preconditions, snapshots, postconditions) = _unpack_pre_snap_posts(
                    wrapper
                )

                resolved_kwargs = kwargs_from_call(
                    param_names=param_names,
                    kwdefaults=kwdefaults,
                    args=args,
                    kwargs=kwargs,
                )

                type_error = _assert_resolved_kwargs_valid(
                    postconditions=postconditions, resolved_kwargs=resolved_kwargs
                )
                if type_error:
                    raise type_error

                violation_error = _assert_preconditions(
                    preconditions=preconditions,
                    resolved_kwargs=resolved_kwargs,
                    func=func,
                )
                if violation_error:
                    raise violation_error

                # Capture the snapshots
                if postconditions and snapshots:
                    resolved_kwargs["OLD"] = _capture_old(
                        snapshots=snapshots, resolved_kwargs=resolved_kwargs, func=func
                    )

                # Ideally, we would catch any exception here and strip the checkers from the traceback.
                # Unfortunately, this can not be done in Python 3, see
                # https://stackoverflow.com/questions/44813333/how-can-i-elide-a-function-wrapper-from-the-traceback-in-python-3
                result = func(*args, **kwargs)

                if postconditions:
                    resolved_kwargs["result"] = result

                    violation_error = _assert_postconditions(
                        postconditions=postconditions,
                        resolved_kwargs=resolved_kwargs,
                        func=func,
                    )
                    if violation_error:
                        raise violation_error

                return result
            finally:
                in_progress.discard(id_func)

    # Copy __doc__ and other properties so that doctests can run
    functools.update_wrapper(wrapper=wrapper, wrapped=func)

    assert not hasattr(
        wrapper, "__preconditions__"
    ), "Expected no preconditions set on a pristine contract checker."
    assert not hasattr(
        wrapper, "__postcondition_snapshots__"
    ), "Expected no postcondition snapshots set on a pristine contract checker."
    assert not hasattr(
        wrapper, "__postconditions__"
    ), "Expected no postconditions set on a pristine contract checker."

    # Precondition is a list of condition groups (i.e. disjunctive normal form):
    # each group consists of AND'ed preconditions, while the groups are OR'ed.
    #
    # This is necessary in order to implement "require else" logic when a class weakens the preconditions of
    # its base class.
    setattr(wrapper, "__preconditions__", [])
    setattr(wrapper, "__postcondition_snapshots__", [])
    setattr(wrapper, "__postconditions__", [])

    return wrapper  # type: ignore


def add_precondition_to_checker(checker: CallableT, contract: Contract) -> None:
    """
    Add the precondition to the function's checker.

    Use :func:`find_checker` to find the checker.
    If it returns ``None``, decorate it with the checker first using :func:`decorate_with_checker`.
    """
    # Add the precondition to the list of preconditions stored at the checker
    assert hasattr(checker, "__preconditions__")
    preconditions = getattr(checker, "__preconditions__")
    assert isinstance(preconditions, list)
    assert len(preconditions) <= 1, (
        "At most a single group of preconditions expected when wrapping with a contract checker. "
        "The preconditions are merged only in the DBC metaclass. "
        "The current number of precondition groups: {}"
    ).format(len(preconditions))

    if len(preconditions) == 0:
        # Create the first group if there is no group so far, i.e. this is the first decorator.
        preconditions.append([])

    preconditions[0].append(contract)


def add_snapshot_to_checker(checker: CallableT, snapshot: Snapshot) -> None:
    """
    Add the snapshot to the function's checker.

    Use :func:`find_checker` to find the checker.
    If it returns ``None``, decorate it with the checker first using :func:`decorate_with_checker`.
    """
    # Add the snapshot to the list of snapshots stored at the checker
    assert hasattr(checker, "__postcondition_snapshots__")

    snapshots = getattr(checker, "__postcondition_snapshots__")
    assert isinstance(snapshots, list)

    for snap in snapshots:
        assert isinstance(snap, Snapshot)
        if snap.name == snapshot.name:
            raise ValueError(
                "There are conflicting snapshots with the name: {!r}".format(snap.name)
            )

    snapshots.append(snapshot)


def add_postcondition_to_checker(checker: CallableT, contract: Contract) -> None:
    """
    Add the postcondition to the function's checker.

    Use :func:`find_checker` to find the checker.
    If it returns ``None``, decorate it with the checker first using :func:`decorate_with_checker`.
    """
    # Add the postcondition to the list of postconditions stored at the checker
    assert hasattr(checker, "__postconditions__")
    assert isinstance(getattr(checker, "__postconditions__"), list)
    getattr(checker, "__postconditions__").append(contract)


def _find_self(
    param_names: List[str], args: Tuple[Any, ...], kwargs: Dict[str, Any]
) -> Any:
    """Find the instance of ``self`` in the arguments."""
    instance_i = None
    try:
        instance_i = param_names.index("self")
    except ValueError:
        pass

    if instance_i is not None and instance_i < len(args):
        return args[instance_i]

    return kwargs["self"]


def _decorate_new_with_invariants(new_func: CallableT) -> CallableT:
    """
    Decorate the ``__new__`` of a class s.t. the invariants are checked on the result.

    This is necessary for optimized classes such as ``namedtuple`` which use ``object.__init__``
    as constructor and do not expect a wrapping around the constructor.
    """
    if _already_decorated_with_invariants(func=new_func):
        return new_func

    # __new__ can not be async in Python, so we don't need to check whether it is sync or async here
    # (as opposed to the function `_decorate_with_invariants`).
    # See this StackOverflow answer: https://stackoverflow.com/a/33134213/1600678.

    def wrapper(*args, **kwargs):  # type: ignore
        """Pass the arguments to __new__ and check invariants on the result."""
        instance = new_func(*args, **kwargs)

        for contract in instance.__class__.__invariants__:
            _assert_invariant(contract=contract, instance=instance)

        return instance

    functools.update_wrapper(wrapper=wrapper, wrapped=new_func)

    setattr(wrapper, "__is_invariant_check__", True)

    return wrapper  # type: ignore


def _decorate_with_invariants(func: CallableT, is_init: bool) -> CallableT:
    """
    Decorate the function ``func`` of the class ``cls`` with invariant checks.

    If the function has been already decorated with invariant checks, the function returns immediately.

    :param func: function to be wrapped
    :param is_init: True if the ``func`` is __init__
    :return: function wrapped with invariant checks
    """
    if _already_decorated_with_invariants(func=func):
        return func

    sign = inspect.signature(func)
    param_names = list(sign.parameters.keys())

    if is_init:

        def wrapper(*args, **kwargs):  # type: ignore
            """Wrap __init__ method of a class by checking the invariants *after* the invocation."""
            try:
                instance = _find_self(param_names=param_names, args=args, kwargs=kwargs)
            except KeyError as err:
                raise KeyError(
                    (
                        "The parameter 'self' could not be found in the call to function {!r}: "
                        "the param names were {!r}, the args were {!r} and kwargs were {!r}"
                    ).format(func, param_names, args, kwargs)
                ) from err

            # We need to disable the invariants check during the constructor.

            # We need to create a new in-progress set if it is None as the ``ContextVar`` does not accept
            # a factory function for the default argument. If we didn't do this, and simply set an empty
            # set as the default, ``ContextVar`` would always point to the same set by copying the default
            # by reference.
            in_progress = _IN_PROGRESS.get()
            if in_progress is None:
                in_progress = set()
                _IN_PROGRESS.set(in_progress)

            id_instance = id(instance)
            in_progress.add(id_instance)

            # ExitStack is not used here due to performance.
            try:
                result = func(*args, **kwargs)

                for contract in instance.__class__.__invariants__:
                    _assert_invariant(contract=contract, instance=instance)

                return result
            finally:
                in_progress.discard(id_instance)

    else:
        # (mristin, 2021-02-16)
        # Admittedly, this branching on sync/async is absolutely monstrous.
        # However, I couldn't find out an easier way to refactor the code so that it supports async.
        # Python expects us to explicitly colour functions as sync/async so we can not just put in an if-statement and
        # introduce an "await".
        #
        # The two wrappers need to be manually maintained in parallel.
        # Whenever you make a change, please inspect manually that both sync and async code exercises equivalent
        # behavior.
        # For example, copy/paste the two blocks of code in separate files and perform a diff.

        if inspect.iscoroutinefunction(func):

            async def wrapper(*args, **kwargs):  # type: ignore
                """Wrap a function of a class by checking the invariants *before* and *after* the invocation."""
                try:
                    instance = _find_self(
                        param_names=param_names, args=args, kwargs=kwargs
                    )
                except KeyError as err:
                    raise KeyError(
                        (
                            "The parameter 'self' could not be found in the call to function {!r}: "
                            "the param names were {!r}, the args were {!r} and kwargs were {!r}"
                        ).format(func, param_names, args, kwargs)
                    ) from err

                # We need to create a new in-progress set if it is None as the ``ContextVar`` does not accept
                # a factory function for the default argument. If we didn't do this, and simply set an empty
                # set as the default, ``ContextVar`` would always point to the same set by copying the default
                # by reference.
                in_progress = _IN_PROGRESS.get()
                if in_progress is None:
                    in_progress = set()
                    _IN_PROGRESS.set(in_progress)

                # The following dunder indicates whether another invariant is currently being checked. If so,
                # we need to suspend any further invariant check to avoid endless recursion.
                id_instance = id(instance)
                if id_instance not in in_progress:
                    in_progress.add(id_instance)
                else:
                    # Do not check any invariants to avoid endless recursion.
                    return await func(*args, **kwargs)

                # ExitStack is not used here due to performance.
                try:
                    for contract in instance.__class__.__invariants__:
                        _assert_invariant(contract=contract, instance=instance)

                    result = await func(*args, **kwargs)

                    for contract in instance.__class__.__invariants__:
                        _assert_invariant(contract=contract, instance=instance)

                    return result
                finally:
                    in_progress.discard(id_instance)

        else:

            def wrapper(*args, **kwargs):  # type: ignore
                """Wrap a function of a class by checking the invariants *before* and *after* the invocation."""
                try:
                    instance = _find_self(
                        param_names=param_names, args=args, kwargs=kwargs
                    )
                except KeyError as err:
                    raise KeyError(
                        (
                            "The parameter 'self' could not be found in the call to function {!r}: "
                            "the param names were {!r}, the args were {!r} and kwargs were {!r}"
                        ).format(func, param_names, args, kwargs)
                    ) from err

                # The following dunder indicates whether another invariant is currently being checked. If so,
                # we need to suspend any further invariant check to avoid endless recursion.

                # We need to create a new in-progress set if it is None as the ``ContextVar`` does not accept
                # a factory function for the default argument. If we didn't do this, and simply set an empty
                # set as the default, ``ContextVar`` would always point to the same set by copying the default
                # by reference.
                in_progress = _IN_PROGRESS.get()
                if in_progress is None:
                    in_progress = set()
                    _IN_PROGRESS.set(in_progress)

                id_instance = id(instance)
                if id_instance not in in_progress:
                    in_progress.add(id_instance)
                else:
                    # Do not check any invariants to avoid endless recursion.
                    return func(*args, **kwargs)

                # ExitStack is not used here due to performance.
                try:
                    for contract in instance.__class__.__invariants__:
                        _assert_invariant(contract=contract, instance=instance)

                    result = func(*args, **kwargs)

                    for contract in instance.__class__.__invariants__:
                        _assert_invariant(contract=contract, instance=instance)

                    return result
                finally:
                    in_progress.discard(id_instance)

    functools.update_wrapper(wrapper=wrapper, wrapped=func)

    setattr(wrapper, "__is_invariant_check__", True)

    return wrapper  # type: ignore


class _DummyClass:
    """Represent a dummy class so that we can infer the type of the slot wrapper."""


_SLOT_WRAPPER_TYPE = type(_DummyClass.__init__)  # pylint: disable=invalid-name


def _already_decorated_with_invariants(func: CallableT) -> bool:
    """Check if the function has been already decorated with an invariant check by going through its decorator stack."""
    already_decorated = False
    for a_decorator in _walk_decorator_stack(func=func):
        if getattr(a_decorator, "__is_invariant_check__", False):
            already_decorated = True
            break

    return already_decorated


def add_invariant_checks(cls: ClassT) -> None:
    """Decorate each of the class functions with invariant checks if not already decorated."""
    # Candidates for the decoration as list of (name, dir() value)
    init_name_func = None  # type: Optional[Tuple[str, Callable[..., None]]]
    names_funcs = []  # type: List[Tuple[str, Callable[..., None]]]
    names_properties = []  # type: List[Tuple[str, property]]

    # Filter out entries in the directory which are certainly not candidates for decoration.
    for name in dir(cls):
        value = getattr(cls, name)

        # __new__ is a special class method (though not marked properly with @classmethod!).
        # We need to ignore __repr__ to prevent endless loops when generating error messages.
        # __getattribute__, __setattr__ and __delattr__ are too invasive and alter the state of the instance.
        # Hence we don't consider them "public".
        if name in [
            "__new__",
            "__repr__",
            "__getattribute__",
            "__setattr__",
            "__delattr__",
        ]:
            continue

        if name == "__init__":
            assert inspect.isfunction(value) or isinstance(
                value, _SLOT_WRAPPER_TYPE
            ), "Expected __init__ to be either a function or a slot wrapper, but got: {}".format(
                type(value)
            )

            init_name_func = (name, value)
            continue

        if (
            not inspect.isfunction(value)
            and not isinstance(value, _SLOT_WRAPPER_TYPE)
            and not isinstance(value, property)
        ):
            continue

        # Ignore "protected"/"private" methods
        if name.startswith("_") and not (name.startswith("__") and name.endswith("__")):
            continue

        if inspect.isfunction(value) or isinstance(value, _SLOT_WRAPPER_TYPE):
            # Ignore class methods
            if getattr(value, "__self__", None) is cls:
                continue

            # Ignore static methods
            # See https://stackoverflow.com/questions/14187973/python3-check-if-method-is-static
            bound_value = inspect.getattr_static(cls, name, None)
            if isinstance(bound_value, staticmethod):
                continue

            names_funcs.append((name, value))

        elif isinstance(value, property):
            names_properties.append((name, value))

        else:
            raise NotImplementedError(
                "Unhandled directory entry of class {} for {}: {}".format(
                    cls, name, value
                )
            )

    if init_name_func:
        name, func = init_name_func

        # We have to distinguish this special case which is used by named
        # tuples and possibly other optimized data structures.
        # In those cases, we have to wrap __new__ instead of __init__.
        if func == object.__init__ and hasattr(cls, "__new__"):
            new_func = getattr(cls, "__new__")
            setattr(cls, "__new__", _decorate_new_with_invariants(new_func))
        else:
            wrapper = _decorate_with_invariants(func=func, is_init=True)
            setattr(cls, name, wrapper)

    for name, func in names_funcs:
        wrapper = _decorate_with_invariants(func=func, is_init=False)
        setattr(cls, name, wrapper)

    for name, prop in names_properties:
        new_prop = property(
            fget=_decorate_with_invariants(func=prop.fget, is_init=False)
            if prop.fget
            else None,
            fset=_decorate_with_invariants(func=prop.fset, is_init=False)
            if prop.fset
            else None,
            fdel=_decorate_with_invariants(func=prop.fdel, is_init=False)
            if prop.fdel
            else None,
            doc=prop.__doc__,
        )
        setattr(cls, name, new_prop)



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
import os
import sys

from setuptools import setup, find_packages

# pylint: disable=redefined-builtin

here = os.path.abspath(os.path.dirname(__file__))  # pylint: disable=invalid-name

with open(os.path.join(here, "README.rst"), encoding="utf-8") as fid:
    long_description = fid.read()  # pylint: disable=invalid-name

# Please keep the meta information in sync with icontract/__init__.py.
#
# (mristin, 2020-10-09) We had to denormalize icontract_meta module (which
# used to be referenced from setup.py and this file) since readthedocs had
# problems with installing icontract through pip on their servers with
# imports in setup.py.
setup(
    name="icontract",
    # Don't forget to update the version in __init__.py and CHANGELOG.rst!
    version="2.6.3",
    description="Provide design-by-contract with informative violation messages.",
    long_description=long_description,
    url="https://github.com/Parquery/icontract",
    author="Marko Ristin",
    author_email="marko@ristin.ch",
    classifiers=[
        # yapf: disable
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
        # yapf: enable
    ],
    license="License :: OSI Approved :: MIT License",
    keywords="design-by-contract precondition postcondition validation",
    packages=find_packages(exclude=["tests*"]),
    install_requires=[
        "asttokens>=2,<3",
        "typing_extensions",
    ],
    extras_require={
        "dev": [
            "pylint==2.17.5",
            "tox>=3.0.0",
            "pydocstyle>=6.3.0,<7",
            "coverage>=4.5.1,<5",
            "docutils>=0.14,<1",
            "pygments>=2.2.0,<3",
            "dpcontracts==0.6.0",
            "tabulate>=0.8.7,<1",
            "py-cpuinfo>=5.0.0,<6",
            "typeguard>=2,<3",
            "astor==0.8.1",
            "numpy>=1,<2",
        ]
        + (["mypy==1.5.1"] if sys.version_info >= (3, 8) else [])
        + (["black==23.9.1"] if sys.version_info >= (3, 8) else [])
        + (["deal==4.23.3"] if sys.version_info >= (3, 8) else [])
        + (["asyncstdlib==3.9.1"] if sys.version_info >= (3, 8) else []),
    },
    py_modules=["icontract"],
    package_data={"icontract": ["py.typed"]},
)



"""Handle representations necessary for informative error messages."""
import ast
import inspect
import re
import reprlib
import sys
import textwrap
import uuid
from typing import (
    Any,
    Mapping,
    MutableMapping,
    Callable,
    List,
    Dict,
    cast,
    Optional,
)  # pylint: disable=unused-import

import asttokens.asttokens

import icontract._recompute
from icontract._types import Contract
from icontract._globals import CallableT

# pylint does not play with typing.Mapping.
# pylint: disable=unsubscriptable-object


def _representable(value: Any) -> bool:
    """
    Check whether we want to represent the value in the error message on contract breach.

    We do not want to represent classes, methods, modules and functions.

    :param value: value related to an AST node
    :return: True if we want to represent it in the violation error
    """
    return (
        not inspect.isclass(value)
        and not inspect.isfunction(value)
        and not inspect.ismethod(value)
        and not inspect.ismodule(value)
        and not inspect.isbuiltin(value)
    )


class Visitor(ast.NodeVisitor):
    """Traverse the abstract syntax tree and collect the representations of the selected nodes."""

    # pylint: disable=invalid-name
    # pylint: disable=missing-docstring

    def __init__(
        self,
        recomputed_values: Mapping[ast.AST, Any],
        variable_lookup: List[Mapping[str, Any]],
        atok: asttokens.asttokens.ASTTokens,
    ) -> None:
        """
        Initialize.

        :param recomputed_values: AST node of a condition function -> value associated with the node
        :param variable_lookup:
            list of lookup tables to look-up the values of the variables, sorted by precedence.
            The visitor needs it here to check whether we overrode a built-in variable (like ``id``).
        :param atok: parsed AST tree and tokens with additional positions in source code

        """
        self._recomputed_values = recomputed_values
        self._variable_lookup = variable_lookup
        self.reprs = dict()  # type: MutableMapping[str, str]
        self._atok = atok

    if sys.version_info >= (3, 6):

        def visit_JoinedStr(self, node: ast.JoinedStr) -> None:
            """Show the whole joined strings without descending into the values."""
            if node in self._recomputed_values:
                value = self._recomputed_values[node]

                if _representable(value=value):
                    text = self._atok.get_text(node)
                    self.reprs[text] = value

    def visit_Name(self, node: ast.Name) -> None:
        """
        Resolve the name from the variable look-up and the built-ins.

        Due to possible branching (e.g., If-expressions), some nodes might lack the recomputed values. These nodes
        are ignored.
        """
        if node in self._recomputed_values:
            value = self._recomputed_values[node]

            # Check if it is a non-built-in
            is_builtin = True
            for lookup in self._variable_lookup:
                if node.id in lookup:
                    is_builtin = False
                    break

            if not is_builtin and _representable(value=value):
                text = self._atok.get_text(node)
                self.reprs[text] = value

        self.generic_visit(node=node)

    def visit_Attribute(self, node: ast.Attribute) -> None:
        """Represent the attribute by dumping its source code."""
        if node in self._recomputed_values:
            value = self._recomputed_values[node]

            if _representable(value=value):
                text = self._atok.get_text(node)
                self.reprs[text] = value

        self.generic_visit(node=node)

    if sys.version_info >= (3, 8):

        def visit_NamedExpr(self, node: ast.NamedExpr) -> Any:
            """Represent the target with the value of the node."""
            if node in self._recomputed_values:
                value = self._recomputed_values[node]

                target = node.target

                if _representable(value=value):
                    self.reprs[target.id] = value

            self.generic_visit(node=node)

    def visit_Call(self, node: ast.Call) -> None:
        """Represent the call by dumping its source code."""
        if node in self._recomputed_values:
            value = self._recomputed_values[node]
            text = self._atok.get_text(node)

            self.reprs[text] = value

        self.generic_visit(node=node)

    def visit_ListComp(self, node: ast.ListComp) -> None:
        """Represent the list comprehension by dumping its source code."""
        if node in self._recomputed_values:
            value = self._recomputed_values[node]
            text = self._atok.get_text(node)

            self.reprs[text] = value

        self.generic_visit(node=node)

    def visit_SetComp(self, node: ast.SetComp) -> None:
        """Represent the set comprehension by dumping its source code."""
        if node in self._recomputed_values:
            value = self._recomputed_values[node]
            text = self._atok.get_text(node)

            self.reprs[text] = value

        self.generic_visit(node=node)

    def visit_DictComp(self, node: ast.DictComp) -> None:
        """Represent the dictionary comprehension by dumping its source code."""
        if node in self._recomputed_values:
            value = self._recomputed_values[node]
            text = self._atok.get_text(node)

            self.reprs[text] = value

        self.generic_visit(node=node)

    def visit_Subscript(self, node: ast.Subscript) -> None:
        """Represent the subscript with its source code."""
        if node in self._recomputed_values:
            value = self._recomputed_values[node]
            text = self._atok.get_text(node)

            self.reprs[text] = value

        self.generic_visit(node=node)


def is_lambda(a_function: CallableT) -> bool:
    """
    Check whether the function is a lambda function.

    >>> def some_func()->bool: return True
    >>> is_lambda(some_func)
    False

    >>> lmbd = lambda x: x > 0
    >>> is_lambda(lmbd)
    True

    :return: True if condition is defined as lambda function
    """
    return a_function.__name__ == "<lambda>"


class ConditionLambdaInspection:
    """Represent the inspection of the condition function given as a lambda."""

    def __init__(self, atok: asttokens.asttokens.ASTTokens, node: ast.Lambda) -> None:
        """
        Initialize.

        :param atok: parsed AST tree and tokens with added positional properties
        :param node: lambda AST node corresponding to the condition
        """
        self.atok = atok
        self.node = node

        text = atok.get_text(node.body)
        assert isinstance(text, str)
        self.text = text


_DECORATOR_RE = re.compile(r"^\s*@[a-zA-Z_]")
_DEF_CLASS_RE = re.compile(r"^\s*(async\s+def|def |class )")


class DecoratorInspection:
    """Represent the inspection of a decorator extracted from a source file and embedded in a dummy dynamic module."""

    def __init__(self, atok: asttokens.asttokens.ASTTokens, node: ast.Call) -> None:
        """
        Initialize.

        :param atok: parsed AST tree and tokens with added positional properties
        :param node: lambda AST node corresponding to the condition
        """
        self.atok = atok
        self.node = node


def inspect_decorator(
    lines: List[str], lineno: int, filename: str
) -> DecoratorInspection:
    """
    Parse the file in which the decorator is called and figure out the corresponding call AST node.

    :param lines: lines of the source file corresponding to the decorator call
    :param lineno: line index (starting with 0) of one of the lines in the decorator call
    :param filename: name of the file where decorator is called
    :return: inspected decorator call
    """
    if lineno < 0 or lineno >= len(lines):
        raise ValueError(
            (
                "Given line number {} of one of the decorator lines "
                "is not within the range [{}, {}) of lines in {}.\n\n"
                "The decorator lines were:\n{}"
            ).format(lineno, 0, len(lines), filename, "\n".join(lines))
        )

    # Go up till a line starts with a decorator
    decorator_lineno = None  # type: Optional[int]
    for i in range(lineno, -1, -1):
        if _DECORATOR_RE.match(lines[i]):
            decorator_lineno = i
            break

    if decorator_lineno is None:
        raise SyntaxError(
            "Decorator corresponding to the line {} could not be found in file {}: {!r}".format(
                lineno + 1, filename, lines[lineno]
            )
        )

    # Find the decorator end -- it's either a function definition, a class definition or another decorator
    decorator_end_lineno = None  # type: Optional[int]
    for i in range(lineno + 1, len(lines)):
        line = lines[i]

        if _DECORATOR_RE.match(line) or _DEF_CLASS_RE.match(line):
            decorator_end_lineno = i
            break

    if decorator_end_lineno is None:
        raise SyntaxError(
            (
                "The next statement following the decorator corresponding to the line {} "
                "could not be found in file {}: {!r}"
            ).format(lineno + 1, filename, lines[lineno])
        )

    decorator_lines = lines[decorator_lineno:decorator_end_lineno]

    # We need to dedent the decorator and add a dummy decorate so that we can parse its text as valid source code.
    decorator_text = textwrap.dedent(
        "".join(decorator_lines)
    ) + "def dummy_{}(): pass".format(uuid.uuid4().hex)

    atok = asttokens.asttokens.ASTTokens(decorator_text, parse=True)

    if not isinstance(atok.tree, ast.Module):
        raise ValueError(
            (
                "Expected the parsed decorator text to live in an AST module. "
                "Are you trying to inspect a condition lambda which was not stated in a decorator? "
                "(This feature is currently unsupported in icontract.) "
                "The decorator was expected at line {} in {}. "
                "The decorator lines under inspection were {}-{}."
            ).format(lineno + 1, filename, decorator_lineno + 1, decorator_end_lineno)
        )

    module_node = atok.tree

    if len(module_node.body) != 1:
        raise ValueError(
            (
                "Expected the module AST of the decorator text to have a single statement. "
                "Are you trying to inspect a condition lambda which was not stated in a decorator? "
                "(This feature is currently unsupported in icontract.) "
                "The decorator was expected at line {} in {}. "
                "The decorator lines under inspection were {}-{}."
            ).format(lineno + 1, filename, decorator_lineno + 1, decorator_end_lineno)
        )

    if not isinstance(module_node.body[0], ast.FunctionDef):
        raise ValueError(
            (
                "Expected the only statement in the AST module corresponding to the decorator text "
                "to be a function definition. "
                "Are you trying to inspect a condition lambda which was not stated in a decorator? "
                "(This feature is currently unsupported in icontract.) "
                "The decorator was expected at line {} in {}. "
                "The decorator lines under inspection were {}-{}."
            ).format(lineno + 1, filename, decorator_lineno + 1, decorator_end_lineno)
        )

    func_def_node = module_node.body[0]

    if len(func_def_node.decorator_list) != 1:
        raise ValueError(
            (
                "Expected the function AST node corresponding to the decorator text to have a single decorator. "
                "Are you trying to inspect a condition lambda which was not stated in a decorator? "
                "(This feature is currently unsupported in icontract.) "
                "The decorator was expected at line {} in {}. "
                "The decorator lines under inspection were {}-{}."
            ).format(lineno + 1, filename, decorator_lineno + 1, decorator_end_lineno)
        )

    if not isinstance(func_def_node.decorator_list[0], ast.Call):
        raise ValueError(
            (
                "Expected the only decorator in the function definition AST node corresponding "
                "to the decorator text to be a call node. "
                "Are you trying to inspect a condition lambda which was not stated in a decorator? "
                "(This feature is currently unsupported in icontract.) "
                "The decorator was expected at line {} in {}. "
                "The decorator lines under inspection were {}-{}."
            ).format(lineno + 1, filename, decorator_lineno + 1, decorator_end_lineno)
        )

    call_node = func_def_node.decorator_list[0]

    return DecoratorInspection(atok=atok, node=call_node)


def find_lambda_condition(
    decorator_inspection: DecoratorInspection,
) -> Optional[ConditionLambdaInspection]:
    """
    Inspect the decorator and extract the condition as lambda.

    If the condition is not given as a lambda function, return None.
    """
    call_node = decorator_inspection.node

    lambda_node = None  # type: Optional[ast.Lambda]

    if len(call_node.args) > 0:
        assert isinstance(call_node.args[0], ast.Lambda), (
            "Expected the first argument to the decorator to be a condition as lambda AST node, "
            "but got: {}"
        ).format(type(call_node.args[0]))

        lambda_node = call_node.args[0]

    elif len(call_node.keywords) > 0:
        for keyword in call_node.keywords:
            if keyword.arg == "condition":
                assert isinstance(
                    keyword.value, ast.Lambda
                ), "Expected lambda node as value of the 'condition' argument to the decorator."

                lambda_node = keyword.value
                break

        assert (
            lambda_node is not None
        ), "Expected to find a keyword AST node with 'condition' arg, but found none"
    else:
        raise AssertionError(
            "Expected a call AST node of a decorator to have either args or keywords, but got: {}".format(
                ast.dump(call_node)
            )
        )

    return ConditionLambdaInspection(atok=decorator_inspection.atok, node=lambda_node)


def inspect_lambda_condition(
    condition: Callable[..., Any]
) -> Optional[ConditionLambdaInspection]:
    """
    Try to extract the source code of the condition as lambda.

    If the condition is not a lambda, returns None.
    """
    if not is_lambda(condition):
        return None

    lines, condition_lineno = inspect.findsource(condition)
    filename = inspect.getsourcefile(condition)
    assert filename is not None

    decorator_inspection = inspect_decorator(
        lines=lines, lineno=condition_lineno, filename=filename
    )

    lambda_inspection = find_lambda_condition(decorator_inspection=decorator_inspection)

    return lambda_inspection


# yapf: disable
def collect_variable_lookup(
        condition: Callable[..., Any],
        resolved_kwargs: Optional[Mapping[str, Any]] = None
) -> List[Mapping[str, Any]]:
    """
    Collect the variable lookups in order of precedence.

    :param condition: contract condition for which we are constructing a variable look-up tables
    :param resolved_kwargs:
        keyword arguments of the original function to be passed over to the condition (where applicable).

        The keyword arguments are added to the variable look-up tables accordingly.
        If ``resolved_kwargs`` is None, no keyword arguments will be added to the variable look-up table.
    """
    # yapf: enable
    variable_lookup = []  # type: List[Mapping[str, Any]]

    ##
    # Condition-specific kwargs
    ##

    if resolved_kwargs is not None:
        variable_lookup.append(resolved_kwargs)

    ##
    # Add closure to the lookup
    ##

    closure_dict = dict()  # type: Dict[str, Any]

    if condition.__closure__ is not None:
        closure_cells = condition.__closure__
        freevars = condition.__code__.co_freevars

        assert len(closure_cells) == len(freevars), \
            "Number of closure cells of a condition function ({}) == number of free vars ({})".format(
                len(closure_cells), len(freevars))

        for cell, freevar in zip(closure_cells, freevars):
            closure_dict[freevar] = cell.cell_contents

    variable_lookup.append(closure_dict)

    ##
    # Add globals to the lookup
    ##

    if condition.__globals__ is not None:
        variable_lookup.append(condition.__globals__)

    return variable_lookup


def repr_values(condition: Callable[..., bool], lambda_inspection: Optional[ConditionLambdaInspection],
                resolved_kwargs: Mapping[str, Any], a_repr: reprlib.Repr) -> List[str]:
    """
    Represent function arguments and frame values in the error message on contract breach.

    :param condition: condition function of the contract
    :param lambda_inspection:
        inspected lambda AST node corresponding to the condition function (None if the condition was not given as a
        lambda function)
    :param resolved_kwargs: arguments put in the function call
    :param a_repr: representation instance that defines how the values are represented.
    :return: list of value representations
    """
    # Hide _ARGS and _KWARGS if they are not part of the condition for better readability
    if '_ARGS' in resolved_kwargs or '_KWARGS' in resolved_kwargs:
        parameters = inspect.signature(condition).parameters
        malleable_kwargs = cast(
            MutableMapping[str, Any],
            resolved_kwargs.copy()  # type: ignore
        )

        if '_ARGS' not in parameters:
            malleable_kwargs.pop('_ARGS', None)

        if '_KWARGS' not in parameters:
            malleable_kwargs.pop('_KWARGS', None)

        selected_kwargs = cast(Mapping[str, Any], malleable_kwargs)
    else:
        selected_kwargs = resolved_kwargs

    # Don't use ``resolved_kwargs`` from this point on.
    # ``selected_kwargs`` is meant to be used instead for better readability of error messages.

    if is_lambda(a_function=condition):
        assert lambda_inspection is not None, "Expected a lambda inspection when given a condition as a lambda function"
    else:
        assert lambda_inspection is None, "Expected no lambda inspection in a condition given as a non-lambda function"

    reprs = None  # type: Optional[MutableMapping[str, Any]]

    if lambda_inspection is not None:
        variable_lookup = collect_variable_lookup(condition=condition, resolved_kwargs=selected_kwargs)

        recompute_visitor = icontract._recompute.Visitor(variable_lookup=variable_lookup)

        recompute_visitor.visit(node=lambda_inspection.node.body)
        recomputed_values = recompute_visitor.recomputed_values

        repr_visitor = Visitor(
            recomputed_values=recomputed_values, variable_lookup=variable_lookup, atok=lambda_inspection.atok)
        repr_visitor.visit(node=lambda_inspection.node.body)

        reprs = repr_visitor.reprs

    # Add original arguments from the call unless they shadow a variable in the re-computation.
    #
    # The condition arguments are often not sufficient to figure out the error. The user usually needs
    # more context which is captured in the remainder of the call arguments.

    if reprs is None:
        reprs = dict()

    for key in sorted(selected_kwargs.keys()):
        val = selected_kwargs[key]
        if key not in reprs and _representable(value=val):
            reprs[key] = val

    parts = []  # type: List[str]

    # We need to sort in order to present the same violation error on repeated violations.
    # Otherwise, the order of the reported arguments may be arbitrary.
    for key in sorted(reprs.keys()):
        value = reprs[key]
        if isinstance(value, icontract._recompute.FirstExceptionInAll):
            writing = ['{} was False, e.g., with'.format(key)]
            for input_name, input_value in value.inputs:
                writing.append('\n')
                writing.append('  {} = {}'.format(input_name, a_repr.repr(input_value)))

            parts.append(''.join(writing))
        else:
            parts.append('{} was {}'.format(key, a_repr.repr(value)))

    return parts


def represent_condition(condition: CallableT) -> str:
    """Represent the condition as a string."""
    lambda_inspection = None  # type: Optional[ConditionLambdaInspection]
    if not is_lambda(a_function=condition):
        condition_repr = condition.__name__
    else:
        # We need to extract the source code corresponding to the decorator since inspect.getsource() is broken with
        # lambdas.
        lambda_inspection = inspect_lambda_condition(condition=condition)
        assert lambda_inspection is not None, "Unexpected no lambda inspection for condition: {}".format(condition)
        condition_repr = lambda_inspection.atok.get_text(lambda_inspection.node)

    return condition_repr


def generate_message(contract: Contract, resolved_kwargs: Mapping[str, Any]) -> str:
    """Generate the message upon contract violation."""
    parts = []  # type: List[str]

    if contract.location is not None:
        parts.append("{}:\n".format(contract.location))

    if contract.description is not None:
        parts.append("{}: ".format(contract.description))

    lambda_inspection = None  # type: Optional[ConditionLambdaInspection]
    if not is_lambda(a_function=contract.condition):
        condition_text = contract.condition.__name__
    else:
        # We need to extract the source code corresponding to the decorator since inspect.getsource() is broken with
        # lambdas.
        lambda_inspection = inspect_lambda_condition(condition=contract.condition)
        assert lambda_inspection is not None, \
            "Unexpected no lambda inspection for condition: {}".format(contract.condition)
        condition_text = lambda_inspection.text

    parts.append(condition_text)

    repr_vals = repr_values(
        condition=contract.condition,
        lambda_inspection=lambda_inspection,
        resolved_kwargs=resolved_kwargs,
        a_repr=contract._a_repr)

    if len(repr_vals) == 0:
        # Do not append anything since no value could be represented as a string.
        # This could appear in case we have, for example, a generator expression as the return value of a lambda.
        pass

    elif len(repr_vals) == 1:
        parts.append(': ')
        parts.append(repr_vals[0])
    else:
        parts.append(':\n')
        parts.append('\n'.join(repr_vals))

    msg = "".join(parts)

    return msg
